package guicop;
import java.util.*;
import java.awt.Rectangle;

public class GCUDefObj extends GCNode {
	public String name;

	public Hashtable<String, GCVar> variables;
	public Hashtable<Integer, GCVar> varsByIndex=null;
	public ArrayList<GCProperty> properties;
	public ArrayList<GCConstraint> constraints;

  public ArrayList<GCComponent> allSolutions;
  public boolean propertiesComputed;
	
	public GCUDefObj(String name) { 
		this.name = name;
		variables = new Hashtable<String, GCVar>();
		properties = new ArrayList<GCProperty>();
		constraints = new ArrayList<GCConstraint>();
    allSolutions = null;
    propertiesComputed = false;
	}

  public boolean matchVariables(int depth) {
    //match all variables
    Enumeration vars = variables.elements();
    int i = 0;
    while(vars.hasMoreElements()){
      GCVar var = (GCVar) vars.nextElement();
      var.index = i;
      var.match(depth+1);

      if(var.matches == null || var.matches.isEmpty()){
        return false;
      }
      var.makeMatchesComposite(i,this);
      i++;
    }
    return true;
  }

  public GCVar getVarByIndex (int varIndex ) {
    if (varsByIndex == null) {
      varsByIndex = new Hashtable<Integer, GCVar>();
      Enumeration vars = variables.elements();
      while(vars.hasMoreElements()){
        GCVar var = (GCVar) vars.nextElement();
        assert(varsByIndex.containsKey(var.index));
        varsByIndex.put(var.index,var);
      }
    }
    GCVar var = varsByIndex.get(varIndex);
    if (var != null) {
      assert(var.index == varIndex);
    }
    return var;
  }

  public ArrayList<GCComponent> computeIntersection(ArrayList<GCComponent> left, ArrayList<GCComponent> right) {
    ArrayList<GCComponent> inter = new ArrayList<GCComponent>();
    for (GCComponent comp1 : left) {
      for (GCComponent comp2 : right) {
        if (comp1.isComposite()){
          GCCompositeComp ccomp = comp1.composite();
          GCComponent comp = ccomp.getIntersection(comp2);
          if(comp != null) {
            inter.add(comp);
          }
        } else if (comp1.isEqual( comp2) ) {
          inter.add(comp1);
        }
      }
    }
    return inter;
  }

 @Override
  public String getPrettyName() {
    return name;
  }

  public boolean matchConstraints(int depth) {
    int i = 0;
    for (GCConstraint constraint : constraints) {
      olog("Constraint " + Integer.toString(i) + " {" );
      ArrayList<GCComponent> constrMatches = constraint.expr.match(depth+1);
      olog("}. Found " + Integer.toString(constrMatches.size()) + " objects matching constraint " + Integer.toString(i) + "." );
      // if this is the first constraint then set matches
      if (i == 0) {
        matches = constrMatches;
        if (matches.isEmpty()){
          // early termination 
          break;
        }
        i++;
        continue;
      } 
      // this is not the first constraint,
      // compute the cumulative intersection
      olog("cumulative constraint intersection: Start { ");
      matches = computeIntersection(matches, constrMatches);
      olog("} done cumulative constraint intersection: found " + Integer.toString(matches.size()) + " objects"); 
      if (matches == null || matches.isEmpty()){
        // early termination, no need to continue
        return false;
      }
      i++;
    }
    return true;
  }

  public void computeProperties(int depth) {
    //if default properties are not implemented
    //  then compute them using the bounding rectangle of 
    //  the components in the bounding rectangle
    boolean xComputed=false;
    boolean yComputed=false;
    boolean wComputed=false;
    boolean hComputed=false;
    boolean textComputed=false;

    for (GCComponent comp : matches) {
      GCCompositeComp composite = comp.composite();

      for (GCProperty prop : properties) {
        if (prop.isText) {
          String textResult= prop.expr.computeText(composite, depth+1);
          if (prop.name.compareToIgnoreCase("text") == 0) {
            textComputed=true;
            composite.text = textResult;
          } else if (prop.name.compareToIgnoreCase("string") == 0) {
            textComputed=true;
            composite.text = textResult;
          }  else {
            composite.textProperties.put(prop.name,textResult);
          }

        } else {
          Integer intResult = prop.expr.computeInt(composite, depth+1);
          if (prop.name.compareToIgnoreCase("x") == 0) {
            xComputed=true;
            composite.x = intResult;
          } else if (prop.name.compareToIgnoreCase("y") == 0) {
            yComputed=true;
            composite.y = intResult;
          } else if (prop.name.compareToIgnoreCase("width") == 0) {
            wComputed=true;
            composite.width = intResult;
          } else if (prop.name.compareToIgnoreCase("height") == 0) {
            hComputed=true;
            composite.height = intResult;
          } else {
            composite.intProperties.put(prop.name,intResult);
          }
        }
      }
      //bounding rect in case properties not defined
      if (!xComputed || !yComputed || !wComputed || !hComputed){
        //if any of x,y,width,and height were not specified, 
        // then they are computed through the bounding 
        // rectangle of the component. 
        Rectangle r = composite.getBoundingRect();
        if(!xComputed) {
          composite.x = (int) r.getX();
        }
        if(!yComputed) {
          composite.y = (int) r.getY();
        }
        if(!wComputed) {
          composite.width = (int) r.getWidth();
        }
        if(!hComputed) {
          composite.height = (int) r.getHeight();
        }
      }
      if (!textComputed) {
        // if text is not specified then 
        // by default text will be the concatenation text of all
        // the components
        composite.text = composite.getConcatenatedText();
      }
    }

    propertiesComputed = true;
  }

	@Override
	public ArrayList<GCComponent> matchExecute(int depth) {
    olog(depth,"matching variables {");

    if (!  matchVariables(depth)) {
      // if at least one of the variables 
      // has no matches, then no need to continue matching constraints.
      olog(depth, "}. done matching variable. none found");
      return matches;
    }
    olog(depth, "}. done matching variables.");

    olog(depth,"matching constraints {");
    //match all constraints
    if (! matchConstraints(depth)) {
      //early termination as well.
      olog(depth,"} done matching constraints. Early termination.");
      return matches;
    }
    olog(depth,"}. done matching constraints. found " + Integer.toString(matches.size()) + " objects.");

    olog(depth,"computing properties. {");

    //compute properties for each component in matches 
    computeProperties(depth);
    olog(depth,"} done computing properties.");
		return matches;
	}

	public String toString()
	{
		return name + "\n" + variables + "\n" + properties + "\n" + constraints + "\n";
	}

	public boolean isValidVariable(String varName)
	{
		if (variables.containsKey(varName) == false)
		{
			elog("variable " + varName + " is not declared in spec " + name);
			return false;
		}
		return true;
	}

	public GCVar getVariable(String varName) {
		return variables.get(varName);		
	}

  class ComputeAllSolutionsRec {
    public ArrayList<GCComponent> matches = new ArrayList<GCComponent>();
    public GCUDefObj obj;
    GCCompositeComp composite=new GCCompositeComp();

    ComputeAllSolutionsRec(GCUDefObj obj) {
      this.obj = obj;
    }

    void doit(int varIndex) {
      GCVar var = null;
      ArrayList<GCComponent> varMatches = null;

      if (varIndex == obj.variables.size()-1){
        //base case of recursion
        //  fill the matches with the components
        var = obj.getVarByIndex(varIndex);
        if (var == null) {
          elog("error getting variable from hash table by order. panic!");
          return;
        }
        varMatches = var.matches;
        for ( GCComponent comp : varMatches) {
          GCComponent comp1 = comp.composite().components[varIndex];
          if (!composite.doesAddingComponentIRespectExclusive(comp1, varIndex, varIndex) ){
            continue;
          }
          composite.components[varIndex] = comp.composite().components[varIndex];
          GCCompositeComp clone = composite.clone();
          matches.add(clone);
          composite.components[varIndex] = GCCompositeComp.joker;
        }
        return;
      }
      // fill up the composite vector at the corresponding
      // position and recursively fill up the other
      // positions
      var = obj.getVarByIndex(varIndex);
      if (var == null) {
        elog("error getting variable from hash table by order. panic!");
        return;
      }
      varMatches = var.matches;
      for ( GCComponent comp : varMatches) {
        GCComponent comp1 = comp.composite().components[varIndex];
        if (!composite.doesAddingComponentIRespectExclusive(comp1, varIndex, varIndex) ){
          continue;
        }
        composite.components[varIndex] = comp1;
        doit(varIndex+1);
      }
    }
  }

  public void printAllSolutions() {
    olog("printing all solutions: {");
    for (GCComponent comp : allSolutions) {
      GCCompositeComp _comp = comp.composite();
      if (_comp==null){
        continue;
      }
      olog(_comp.componentsToString());
    }
    olog("}done printing all solutions");
    
  }

  public ArrayList<GCComponent> computeAllSolutions() {
    //compute the product of all matches to variables
    if (allSolutions!= null) {
      // if solutions are all cached, return the cached solutions
      return allSolutions;
    }
    olog("Object " + name + ": computing all solutions with " + Integer.toString(variables.size()) + " variables. This may take some time." );
    //recursively compute the solutions
    ComputeAllSolutionsRec compute = new ComputeAllSolutionsRec(this);
    compute.doit(0);
    olog("Object " + name + ": done computing all solutions.");
    // cache the solutions
    allSolutions = compute.matches;
    printAllSolutions();
    // return them
    return allSolutions;
  }
}
